#uses "fwConfigs/fwDPELock.ctl"
#uses "fwUnitTestComponentAsserts.ctl"

const string test_dp = "fwDPELock_test_tryLock";

void assertLocked(string dpe, string config="_original") {
	bool isLocked;
	dyn_string lockDetails;
	dyn_string exceptionInfo;
	
	fwDPELock_getConfigLock(dpe, config, isLocked, lockDetails, exceptionInfo);
	assertTrue(isLocked, "DPE is not locked: " + dpe + ":" + config);
	assertEmpty(exceptionInfo, "Exceptions have been raised");
}

void assertNotLocked(string dpe, string config="_original") {
	bool isLocked;
	dyn_string lockDetails;
	dyn_string exceptionInfo;
	
	fwDPELock_getConfigLock(dpe, config, isLocked, lockDetails, exceptionInfo);
	assertFalse(isLocked, "DPE is locked: " + dpe + ":" + config);
	assertEmpty(exceptionInfo, "Exceptions have been raised");
}

void test_fwDPELock_tryLock_setupSuite() {
	if(dpExists(test_dp))
		dpDelete(test_dp);
		
	dpCreate(test_dp, "ExampleDP_Int");
	dpSetWait(test_dp + ".:_general.._type", DPCONFIG_GENERAL);
}

void test_fwDPELock_tryLock_teardownSuite() {
	if(dpExists(test_dp))
		dpDelete(test_dp);
}

/**
 * 
 */
void test_fwDPELock_tryLock_dpDoesNotExists() {
	dyn_string exceptionInfo;
	bool locked = fwDPELock_tryLock(test_dp + "notExists.", exceptionInfo);
	
	assertFalse(locked, "Lock inexistant config?!");
	assertTrue(dynlen(exceptionInfo) > 0, "Exceptions have not been raised");
	assertNotLocked(test_dp + ".");
}

/**
 * 
 */
void test_fwDPELock_tryLock_dpConfigBlank() {
	dyn_string exceptionInfo;
	bool locked = fwDPELock_tryLock(test_dp + ".", exceptionInfo);
	
	assertTrue(locked, "Could not lock config");
	assertEmpty(exceptionInfo, "Exceptions have been raised");
	assertLocked(test_dp + ".");
}

/**
 * 
 */
void test_fwDPELock_tryLock_dpConfigSpecified() {
	dyn_string exceptionInfo;
	bool locked = fwDPELock_tryLock(test_dp + ".:_general", exceptionInfo);
	
	assertTrue(locked, "Could not lock config _general");
	assertEmpty(exceptionInfo, "Exceptions have been raised");
	assertLocked(test_dp + ".", "_general");
}

/**
 * 
 */
/*void test_fwDPELock_tryLock_timeoutSpecifiedFails() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_timeoutForever_successAfterDelay() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_timeoutImmediateFails() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_timeoutDefaultFails() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_alreadyHasLock() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_lockedByOtherLocal() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_lockedByRemoteFails() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_lockRemoteSuccess() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_lockRemoteFails() {
//TODO: implement test
}*/

/**
 * 
 */
/*void test_fwDPELock_tryLock_lockRedu() {
//TODO: implement test
}*/

